from csle_tolerance.dao.intrusion_response_cmdp_config import (
    IntrusionResponseCmdpConfig,
)
import pytest_mock


class TestIntrusionResponseCmdpConfigSuite:
    """
    Test suite for intrusion_response_cmdp_config.py
    """

    def test__init__(self) -> None:
        """
        Tests the function of initializing the DTO
        """
        p_a = 0.1
        p_c = 0.2
        p_u = 0.3
        s_max = 10
        transition_tensor = [[[0.1, 0.2], [0.3, 0.4]], [[0.5, 0.6], [0.7, 0.8]]]
        cost_tensor = [0.5, 0.6]
        negate_costs = True
        seed = 123
        states = [0, 1]
        actions = [0, 1]
        initial_state = 0
        constraint_cost_tensor = [0.7, 0.8]
        f = 5
        epsilon_a = 0.9
        simulation_env_name = "SimulationEnv"
        gym_env_name = "GymEnv"
        discount_factor = 0.95
        dto = IntrusionResponseCmdpConfig(
            p_a,
            p_c,
            p_u,
            s_max,
            transition_tensor,
            cost_tensor,
            negate_costs,
            seed,
            states,
            actions,
            initial_state,
            constraint_cost_tensor,
            f,
            epsilon_a,
            simulation_env_name,
            gym_env_name,
            discount_factor,
        )
        assert dto.p_a == p_a
        assert dto.p_c == p_c
        assert dto.p_u == p_u
        assert dto.s_max == s_max
        assert dto.transition_tensor == transition_tensor
        assert dto.cost_tensor == cost_tensor
        assert dto.negate_costs == negate_costs
        assert dto.seed == seed
        assert dto.states == states
        assert dto.actions == actions
        assert dto.initial_state == initial_state
        assert dto.constraint_cost_tensor == constraint_cost_tensor
        assert dto.f == f
        assert dto.epsilon_a == epsilon_a
        assert dto.simulation_env_name == simulation_env_name
        assert dto.gym_env_name == gym_env_name
        assert dto.discount_factor == discount_factor

    def test__str__(self) -> None:
        """
        Tests the function of returning a string representation of the DTO
        """
        p_a = 0.1
        p_c = 0.2
        p_u = 0.3
        s_max = 10
        transition_tensor = [[[0.1, 0.2], [0.3, 0.4]], [[0.5, 0.6], [0.7, 0.8]]]
        cost_tensor = [0.5, 0.6]
        negate_costs = True
        seed = 123
        states = [0, 1]
        actions = [0, 1]
        initial_state = 0
        constraint_cost_tensor = [0.7, 0.8]
        f = 5
        epsilon_a = 0.9
        simulation_env_name = "SimulationEnv"
        gym_env_name = "GymEnv"
        discount_factor = 0.95
        dto = IntrusionResponseCmdpConfig(
            p_a,
            p_c,
            p_u,
            s_max,
            transition_tensor,
            cost_tensor,
            negate_costs,
            seed,
            states,
            actions,
            initial_state,
            constraint_cost_tensor,
            f,
            epsilon_a,
            simulation_env_name,
            gym_env_name,
            discount_factor,
        )
        expected = (
            f"p_a: {p_a}, p_c: {p_c}, p_u: {p_u}, s_max: {s_max}, "
            f"transition_tensor: {transition_tensor}, cost_tensor: {cost_tensor}, seed: {seed}, "
            f"negate_costs: {negate_costs}, states: {states}, actions: {actions}, "
            f"initial state: {initial_state}, constraint_cost_tensor: {constraint_cost_tensor}, "
            f"f: {f}, epsilon_a: {epsilon_a}, simulation_env_name: {simulation_env_name}, "
            f"gym_env_name: {gym_env_name}, discount_factor: {discount_factor}"
        )
        assert dto.__str__() == expected

    def test_from_dict(self) -> None:
        """
        Tests the function of converting a dict representation to an instance
        """
        dto_dict = {
            "p_a": 0.1,
            "p_c": 0.2,
            "p_u": 0.3,
            "s_max": 10,
            "transition_tensor": [[[0.1, 0.2], [0.3, 0.4]], [[0.5, 0.6], [0.7, 0.8]]],
            "cost_tensor": [0.5, 0.6],
            "negate_costs": True,
            "seed": 123,
            "states": [0, 1],
            "actions": [0, 1],
            "initial_state": 0,
            "constraint_cost_tensor": [0.7, 0.8],
            "f": 5,
            "epsilon_a": 0.9,
            "simulation_env_name": "SimulationEnv",
            "gym_env_name": "GymEnv",
            "discount_factor": 0.95,
        }
        dto = IntrusionResponseCmdpConfig.from_dict(dto_dict)
        assert dto.p_a == dto_dict["p_a"]
        assert dto.p_c == dto_dict["p_c"]
        assert dto.p_u == dto_dict["p_u"]
        assert dto.s_max == dto_dict["s_max"]
        assert dto.transition_tensor == dto_dict["transition_tensor"]
        assert dto.cost_tensor == dto_dict["cost_tensor"]
        assert dto.negate_costs == dto_dict["negate_costs"]
        assert dto.seed == dto_dict["seed"]
        assert dto.states == dto_dict["states"]
        assert dto.actions == dto_dict["actions"]
        assert dto.initial_state == dto_dict["initial_state"]
        assert dto.constraint_cost_tensor == dto_dict["constraint_cost_tensor"]
        assert dto.f == dto_dict["f"]
        assert dto.epsilon_a == dto_dict["epsilon_a"]
        assert dto.simulation_env_name == dto_dict["simulation_env_name"]
        assert dto.gym_env_name == dto_dict["gym_env_name"]
        assert dto.discount_factor == dto_dict["discount_factor"]

    def test_to_dict(self) -> None:
        """
        Tests the function of getting a dict representation of the object
        """
        p_a = 0.1
        p_c = 0.2
        p_u = 0.3
        s_max = 10
        transition_tensor = [[[0.1, 0.2], [0.3, 0.4]], [[0.5, 0.6], [0.7, 0.8]]]
        cost_tensor = [0.5, 0.6]
        negate_costs = True
        seed = 123
        states = [0, 1]
        actions = [0, 1]
        initial_state = 0
        constraint_cost_tensor = [0.7, 0.8]
        f = 5
        epsilon_a = 0.9
        simulation_env_name = "SimulationEnv"
        gym_env_name = "GymEnv"
        discount_factor = 0.95
        dto = IntrusionResponseCmdpConfig(
            p_a,
            p_c,
            p_u,
            s_max,
            transition_tensor,
            cost_tensor,
            negate_costs,
            seed,
            states,
            actions,
            initial_state,
            constraint_cost_tensor,
            f,
            epsilon_a,
            simulation_env_name,
            gym_env_name,
            discount_factor,
        )
        expected = {
            "p_a": 0.1,
            "p_c": 0.2,
            "p_u": 0.3,
            "s_max": 10,
            "transition_tensor": [[[0.1, 0.2], [0.3, 0.4]], [[0.5, 0.6], [0.7, 0.8]]],
            "cost_tensor": [0.5, 0.6],
            "negate_costs": True,
            "seed": 123,
            "states": [0, 1],
            "actions": [0, 1],
            "initial_state": 0,
            "constraint_cost_tensor": [0.7, 0.8],
            "f": 5,
            "epsilon_a": 0.9,
            "simulation_env_name": "SimulationEnv",
            "gym_env_name": "GymEnv",
            "discount_factor": 0.95,
        }
        assert dto.to_dict() == expected

    def test_from_json_file(self, mocker: pytest_mock.MockFixture) -> None:
        """
        tests the function of reading a json file and converting it to a DTO
        """
        p_a = 0.1
        p_c = 0.2
        p_u = 0.3
        s_max = 10
        transition_tensor = [[[0.1, 0.2], [0.3, 0.4]], [[0.5, 0.6], [0.7, 0.8]]]
        cost_tensor = [0.5, 0.6]
        negate_costs = True
        seed = 123
        states = [0, 1]
        actions = [0, 1]
        initial_state = 0
        constraint_cost_tensor = [0.7, 0.8]
        f = 5
        epsilon_a = 0.9
        simulation_env_name = "SimulationEnv"
        gym_env_name = "GymEnv"
        discount_factor = 0.95
        config = IntrusionResponseCmdpConfig(
            p_a,
            p_c,
            p_u,
            s_max,
            transition_tensor,
            cost_tensor,
            negate_costs,
            seed,
            states,
            actions,
            initial_state,
            constraint_cost_tensor,
            f,
            epsilon_a,
            simulation_env_name,
            gym_env_name,
            discount_factor,
        )
        mocker.patch(
            "io.open", side_effect=mocker.mock_open(read_data=config.to_json_str())
        )
        dto = IntrusionResponseCmdpConfig.from_json_file("path")
        assert dto.p_a == p_a
        assert dto.p_c == p_c
        assert dto.p_u == p_u
        assert dto.s_max == s_max
        assert dto.transition_tensor == transition_tensor
        assert dto.cost_tensor == cost_tensor
        assert dto.negate_costs == negate_costs
        assert dto.seed == seed
        assert dto.states == states
        assert dto.actions == actions
        assert dto.initial_state == initial_state
        assert dto.constraint_cost_tensor == constraint_cost_tensor
        assert dto.f == f
        assert dto.epsilon_a == epsilon_a
        assert dto.simulation_env_name == simulation_env_name
        assert dto.gym_env_name == gym_env_name
        assert dto.discount_factor == discount_factor



from csle_tolerance.util.intrusion_recovery_pomdp_util import IntrusionRecoveryPomdpUtil
from csle_tolerance.dao.intrusion_recovery_pomdp_config import (
    IntrusionRecoveryPomdpConfig,
)
from csle_tolerance.dao.intrusion_recovery_game_config import (
    IntrusionRecoveryGameConfig,
)
import numpy as np
import pytest


class TestIntrusionTolerancePomdpSuite:
    """
    Test suite for intrusion_recovery_pomdp_util.py
    """

    def test__state_space(self) -> None:
        """
        Tests the state space of the POMDP

        :return: None
        """
        assert (isinstance(item, int) for item in IntrusionRecoveryPomdpUtil.state_space())
        assert IntrusionRecoveryPomdpUtil.state_space() is not None
        assert IntrusionRecoveryPomdpUtil.state_space() == [0, 1, 2]

    def test_initial_belief(self) -> None:
        """
        Tests the initial_belief function

        :return: None
        """
        assert sum(IntrusionRecoveryPomdpUtil.initial_belief()) == 1

    def test_action_space(self) -> None:
        """
        Tests the action space of the POMDP

        :return: None
        """
        assert (isinstance(item, int) for item in IntrusionRecoveryPomdpUtil.action_space())
        assert IntrusionRecoveryPomdpUtil.action_space() is not None
        assert IntrusionRecoveryPomdpUtil.action_space() == [0, 1]

    def test_observation_space(self) -> None:
        """
        Tests the observation space of the POMDP

        :return: None
        """
        num_observation = 3
        expected = [0, 1, 2]
        assert IntrusionRecoveryPomdpUtil.observation_space(num_observation) == expected

    def test_cost_function(self) -> None:
        """
        Tests the cost function of the POMDP

        :return: None
        """
        s = 1
        a = 0
        eta = 0.5
        negate = False
        assert IntrusionRecoveryPomdpUtil.cost_function(s, a, eta, negate) == 0.5

    def test_cost_tensor(self) -> None:
        """
        Tests the function of creating a tensor with the costs of the POMDP

        :return: None
        """
        eta = 0.5
        states = [0, 1]
        actions = [0]
        negate = False
        expected = [[0, 0.5]]
        assert IntrusionRecoveryPomdpUtil.cost_tensor(eta, states, actions, negate) == expected

    def test_observation_function(self) -> None:
        """
        Tests the observation function of the POMDP

        :return: None
        """
        s = 1
        o = 1
        num_observations = 2
        assert round(IntrusionRecoveryPomdpUtil.observation_function(s, o, num_observations), 1)

    def test_observation_tensor(self) -> None:
        """
        Tests the function of creating a tensor with observation probabilities

        :return: None
        """
        states = [0, 1]
        observations = [0, 1]
        expected = [[0.8, 0.2], [0.4, 0.6]]
        obs_tensor = IntrusionRecoveryPomdpUtil.observation_tensor(states, observations)
        for i in range(len(obs_tensor)):
            for j in range(len(obs_tensor[i])):
                obs_tensor[i][j] = round(obs_tensor[i][j], 1)
            assert sum(obs_tensor[i]) == 1
        assert obs_tensor == expected

    def test_transition_function(self) -> None:
        """
        Tests the transition function of the POMDP

        :return: None
        """
        s = 0
        s_prime = 1
        a = 0
        p_a = 0.2
        p_c_1 = 0.1
        p_c_2 = 0.2
        p_u = 0.5
        assert (round(IntrusionRecoveryPomdpUtil.transition_function(s, s_prime, a, p_a, p_c_1, p_c_2, p_u), 1) == 0.2)

    def test_transition_function_game(self) -> None:
        """
        Tests the transition function of the POSG

        :return: None
        """
        s = 0
        s_prime = 1
        a1 = 0
        a2 = 1
        p_a = 0.2
        p_c_1 = 0.1
        assert (round(IntrusionRecoveryPomdpUtil.transition_function_game(s, s_prime, a1, a2, p_a, p_c_1), 2) == 0.18)

    def test_transition_tensor(self) -> None:
        """
        Tests the function of creating  a tensor with the transition probabilities of the POMDP

        :return: None
        """
        states = [0, 1, 2]
        actions = [0]
        p_a = 0.2
        p_c_1 = 0.1
        p_c_2 = 0.2
        p_u = 0.5
        expected = [[[0.7, 0.2, 0.1], [0.4, 0.4, 0.2], [0, 0, 1.0]]]
        transition_tensor = IntrusionRecoveryPomdpUtil.transition_tensor(states, actions, p_a, p_c_1, p_c_2, p_u)
        for i in range(len(transition_tensor)):
            for j in range(len(transition_tensor[i])):
                for k in range(len(transition_tensor[i][j])):
                    transition_tensor[i][j][k] = round(transition_tensor[i][j][k], 1)
        assert transition_tensor == expected
        states = [0, 1]
        with pytest.raises(AssertionError):
            IntrusionRecoveryPomdpUtil.transition_tensor(states, actions, p_a, p_c_1, p_c_2, p_u)

    def test_transition_tensor_game(self) -> None:
        """
        Tests the function of creating a tensor with the transition probabilities of the POSG

        :return: None
        """
        states = [0, 1, 2]
        defender_actions = [0, 1]
        attacker_actions = [0, 1]
        p_a = 0.5
        p_c_1 = 0.3
        result = IntrusionRecoveryPomdpUtil.transition_tensor_game(states, defender_actions, attacker_actions, p_a,
                                                                   p_c_1)
        assert len(result) == len(defender_actions)
        assert all(len(a1) == len(attacker_actions) for a1 in result)
        assert all(len(a2) == len(states) for a1 in result for a2 in a1)
        assert all(len(s) == len(states) for a1 in result for a2 in a1 for s in a2)
        assert result[0][1][0][0] == (1 - p_a) * (1 - p_c_1)
        assert result[1][0][1][1] == 0
        assert result[1][1][2][2] == 1.0
        assert result[0][1][0][1] == (1 - p_c_1) * p_a
        assert result[0][0][0][2] == p_c_1

    def test_sample_initial_state(self) -> None:
        """
        Tests the function of sampling the initial state

        :return: None
        """
        b1 = [0.2, 0.8]
        assert isinstance(IntrusionRecoveryPomdpUtil.sample_initial_state(b1), int)
        assert IntrusionRecoveryPomdpUtil.sample_initial_state(b1) <= len(b1)
        assert IntrusionRecoveryPomdpUtil.sample_initial_state(b1) >= 0
        b1 = [1.0, 0]
        assert IntrusionRecoveryPomdpUtil.sample_initial_state(b1) == 0
        b1 = [0.0, 1.0]
        assert IntrusionRecoveryPomdpUtil.sample_initial_state(b1) == 1

    def test_sampe_next_observation(self) -> None:
        """
        Tests the function of sampling the next observation

        :return: None
        """
        observation_tensor = [[0.8, 0.2], [0.4, 0.6]]
        s_prime = 1
        observations = [0, 1]
        assert isinstance(IntrusionRecoveryPomdpUtil.sample_next_observation(observation_tensor, s_prime, observations),
                          int)

    def test_bayes_filter(self) -> None:
        """
        Tests the function of a bayesian filter to computer b[s_prime] of the POMDP

        :return: None
        """
        s_prime = 1
        o = 0
        a = 0
        b = [0.2, 0.8]
        states = [0, 1]
        observations = [0, 1]
        observation_tensor = [[0.8, 0.2], [0.4, 0.6]]
        transition_tensor = [[[0.6, 0.4], [0.1, 0.9]]]
        b_prime_s_prime = 0.7
        assert (round(IntrusionRecoveryPomdpUtil.bayes_filter(s_prime, o, a, b, states, observations,
                                                              observation_tensor, transition_tensor), 1)
                == b_prime_s_prime)

    def test_p_o_given_b_a1_a2(self) -> None:
        """
        Tests the function of computing P[o|a,b] of the POMDP

        :return: None
        """
        o = 0
        b = [0.2, 0.8]
        a = 0
        states = [0, 1]
        observation_tensor = [[0.8, 0.2], [0.4, 0.6]]
        transition_tensor = [[[0.6, 0.4], [0.1, 0.9]]]
        expected = 0.5
        assert (round(IntrusionRecoveryPomdpUtil.p_o_given_b_a1_a2(o, b, a, states, transition_tensor,
                                                                   observation_tensor), 1) == expected)

    def test_next_belief(self) -> None:
        """
        Tests the funtion of computing the next belief using a Bayesian filter

        :return: None
        """
        o = 0
        a = 0
        b = [0.2, 0.8]
        states = [0, 1]
        observations = [0, 1]
        observation_tensor = [[0.8, 0.2], [0.4, 0.6]]
        transition_tensor = [[[0.3, 0.7], [0.6, 0.4]]]
        assert (round(sum(IntrusionRecoveryPomdpUtil.next_belief(o, a, b, states, observations, observation_tensor,
                                                                 transition_tensor)), 1) == 1)

    def test_pomdp_solver_file(self) -> None:
        """
        Tests the function of getting the POMDP environment specification

        :return: None
        """

        assert (IntrusionRecoveryPomdpUtil.pomdp_solver_file(
            IntrusionRecoveryPomdpConfig(eta=0.1, p_a=0.2, p_c_1=0.2, p_c_2=0.3, p_u=0.3, BTR=1, negate_costs=True,
                                         seed=1, discount_factor=0.5, states=[0, 1], actions=[0], observations=[0, 1],
                                         cost_tensor=[[0.1, 0.5], [0.5, 0.6]],
                                         observation_tensor=[[0.8, 0.2], [0.4, 0.6]],
                                         transition_tensor=[[[0.8, 0.2], [0.6, 0.4]]], b1=[0.3, 0.7], T=3,
                                         simulation_env_name="env", gym_env_name="gym", max_horizon=np.inf))
                is not None)

    def test_sample_next_state_game(self) -> None:
        """
        Tests the function of sampling the next observation

        :return: None
        """
        np.random.seed(40)
        transition_tensor = [
            [
                [[0.1, 0.9], [0.7, 0.3], [0.5, 0.5]],
                [[0.2, 0.8], [0.6, 0.4], [0.4, 0.6]],
            ],
            [
                [[0.3, 0.7], [0.8, 0.2], [0.6, 0.4]],
                [[0.4, 0.6], [0.5, 0.5], [0.3, 0.7]],
            ],
        ]

        s = 0
        a1 = 0
        a2 = 0

        count = [0, 0]
        for _ in range(1000):
            s_prime = IntrusionRecoveryPomdpUtil.sample_next_state_game(
                transition_tensor, s, a1, a2
            )
            count[s_prime] += 1
        assert 850 <= count[1] <= 950
        assert 50 <= count[0] <= 150

    def test_generate_transitions(self) -> None:
        """
        Tests the function of generating the transition rows of the POSG config file of HSVI

        :return: None
        """
        dto = IntrusionRecoveryGameConfig(
            eta=0.5,
            p_a=0.8,
            p_c_1=0.1,
            BTR=10,
            negate_costs=True,
            seed=123,
            discount_factor=0.9,
            states=[0, 1, 2],
            actions=[0, 1],
            observations=[0, 1],
            cost_tensor=[[1, 2], [3, 4]],
            observation_tensor=[[0.6, 0.4], [0.5, 0.5]],
            transition_tensor=[
                [
                    [
                        [0.1, 0.2, 0.7],
                        [0.3, 0.4, 0.3],
                        [0.2, 0.2, 0.6],
                    ],
                    [
                        [0.2, 0.3, 0.5],
                        [0.4, 0.3, 0.3],
                        [0.3, 0.3, 0.4],
                    ],
                ],
                [
                    [
                        [0.5, 0.3, 0.2],
                        [0.3, 0.5, 0.2],
                        [0.4, 0.3, 0.3],
                    ],
                    [
                        [0.3, 0.4, 0.3],
                        [0.5, 0.4, 0.1],
                        [0.2, 0.5, 0.3],
                    ],
                ],
            ],
            b1=[0.1, 0.9],
            T=100,
            simulation_env_name="sim_env",
            gym_env_name="gym_env",
            max_horizon=1000,
        )
        assert IntrusionRecoveryPomdpUtil.generate_transitions(dto)[0] == "0 0 0 0 0 0.06"

    def test_generate_rewards(self) -> None:
        """
        Tests the function of generating the reward rows of the POSG config file of HSVI

        :return: None
        """
        dto = IntrusionRecoveryGameConfig(
            eta=0.5,
            p_a=0.8,
            p_c_1=0.1,
            BTR=10,
            negate_costs=True,
            seed=123,
            discount_factor=0.9,
            states=[0, 1, 2],
            actions=[0, 1],
            observations=[0, 1],
            cost_tensor=[[1, 2, 3], [4, 5, 6]],
            observation_tensor=[[0.6, 0.4], [0.5, 0.5]],
            transition_tensor=[
                [
                    [
                        [0.1, 0.2, 0.7],
                        [0.3, 0.4, 0.3],
                        [0.2, 0.2, 0.6],
                    ],
                    [
                        [0.2, 0.3, 0.5],
                        [0.4, 0.3, 0.3],
                        [0.3, 0.3, 0.4],
                    ],
                ],
                [
                    [
                        [0.5, 0.3, 0.2],
                        [0.3, 0.5, 0.2],
                        [0.4, 0.3, 0.3],
                    ],
                    [
                        [0.3, 0.4, 0.3],
                        [0.5, 0.4, 0.1],
                        [0.2, 0.5, 0.3],
                    ],
                ],
            ],
            b1=[0.1, 0.9],
            T=100,
            simulation_env_name="sim_env",
            gym_env_name="gym_env",
            max_horizon=1000,
        )
        assert IntrusionRecoveryPomdpUtil.generate_rewards(dto)[0] == "0 0 0 -1"

    def test_generate_os_posg_game_file(self) -> None:
        """
        Tests the generate_os_posg_game function

        :return: None
        """

        states = [0, 1, 2]
        actions = [0, 1]
        observations = [0, 1]

        transition_tensor = [
            [
                [
                    [0.1, 0.2, 0.7],
                    [0.3, 0.4, 0.3],
                    [0.2, 0.2, 0.6],
                ],
                [
                    [0.2, 0.3, 0.5],
                    [0.4, 0.3, 0.3],
                    [0.3, 0.3, 0.4],
                ],
            ],
            [
                [
                    [0.5, 0.3, 0.2],
                    [0.3, 0.5, 0.2],
                    [0.4, 0.3, 0.3],
                ],
                [
                    [0.3, 0.4, 0.3],
                    [0.5, 0.4, 0.1],
                    [0.2, 0.5, 0.3],
                ],
            ],
        ]

        observation_tensor = [
            [0.6, 0.4],
            [0.5, 0.5],
        ]

        cost_tensor = [
            [1.0, 2.0, 3.0],
            [4.0, 5.0, 6.0],
        ]

        game_config = IntrusionRecoveryGameConfig(
            eta=0.5,
            p_a=0.8,
            p_c_1=0.1,
            BTR=10,
            negate_costs=True,
            seed=123,
            discount_factor=0.9,
            states=states,
            actions=actions,
            observations=observations,
            cost_tensor=cost_tensor,
            observation_tensor=observation_tensor,
            transition_tensor=transition_tensor,
            b1=[0.1, 0.9],
            T=100,
            simulation_env_name="sim_env",
            gym_env_name="gym_env",
            max_horizon=1000,
        )

        game_file_str = IntrusionRecoveryPomdpUtil.generate_os_posg_game_file(
            game_config
        )

        expected_game_description = "3 1 2 2 2 72 12 0.9"
        expected_state_descriptions = ["0 0", "1 0", "2 0"]
        expected_player_1_actions = ["WAIT", "RECOVER"]
        expected_player_2_actions = ["FALSEALARM", "ATTACK"]
        expected_obs_descriptions = ["o_0", "o_1"]
        expected_player_2_legal_actions = ["0 1", "0 1", "0 1"]
        expected_player_1_legal_actions = ["0 1"]

        output_lines = game_file_str.split("\n")

        assert (output_lines[0] == expected_game_description), f"Game description mismatch: {output_lines[0]}"
        assert (output_lines[1:4] == expected_state_descriptions), f"State descriptions mismatch: {output_lines[1:4]}"
        assert (output_lines[4:6] == expected_player_1_actions), f"Player 1 actions mismatch: {output_lines[4:6]}"
        assert (output_lines[6:8] == expected_player_2_actions), f"Player 2 actions mismatch: {output_lines[6:8]}"
        assert (output_lines[8:10] == expected_obs_descriptions), \
            f"Observation descriptions mismatch: {output_lines[8:10]}"
        assert (output_lines[10:13] == expected_player_2_legal_actions), \
            f"Player 2 legal actions mismatch: {output_lines[10:13]}"
        assert (output_lines[13:14] == expected_player_1_legal_actions), \
            f"Player 1 legal actions mismatch: {output_lines[13:14]}"



from csle_tolerance.util.general_util import GeneralUtil
import numpy as np
import pytest_mock


class TestIntrusionToleranceGeneralSuite:
    """
    Test suite for general_util.py
    """

    def test_threshold_probability(self) -> None:
        """
        Tests the functuon of returning the probability of taking an action given a belief and a threshold

        :return: None
        """
        b1 = 0.5
        threshold = 0.3
        expected = 1
        assert (
            round(GeneralUtil.threshold_probability(b1, threshold, -20), 1) == expected
        )

    def test_sigmoid(self) -> None:
        """
        Tests the sigmoid function

        :return: None
        """
        x = 1
        expected = 0.7
        assert round(GeneralUtil.sigmoid(x), 1) == expected

    def test_inverse_sigmoid(self) -> None:
        """
        Tests the inverse sigmoid function

        :return: None
        """
        x = 0.1
        expected = -2.2
        assert round(GeneralUtil.inverse_sigmoid(x), 1) == expected

    def test_sample_next_state(self) -> None:
        """
        Tests the function of sampling the next state of a MDP or POMDP

        :return: None
        """
        transition_tensor = [[[0.6, 0.4], [0.4, 0.6]]]
        s = 0
        a = 0
        states = [0, 1]
        expected = np.arange(0, len(states))
        assert (
            GeneralUtil.sample_next_state(transition_tensor, s, a, states) in expected
        )

    def test_register_envs(self, mocker: pytest_mock.MockFixture) -> None:
        """
        Tests the utility method for registering Gymnasum environments

        :return: None
        """
        mocked_register = mocker.MagicMock()
        mocker.patch(
            "gymnasium.envs.registration.register", return_value=mocked_register
        )
        mocked_register.configure_mock(**{"__enter__.return_value": None})
        return_value = GeneralUtil.register_envs()
        assert return_value is None



from typing import Tuple, Any, Dict, Union, List
import math
import time
import gymnasium as gym
import numpy as np
from csle_common.dao.simulation_config.base_env import BaseEnv
from csle_common.dao.simulation_config.simulation_trace import SimulationTrace
import csle_common.constants.constants as constants
from csle_tolerance.dao.intrusion_response_cmdp_config import IntrusionResponseCmdpConfig
from csle_tolerance.util.general_util import GeneralUtil
import csle_tolerance.constants.constants as env_constants


class IntrusionResponseCmdpEnv(BaseEnv):
    """
    Gym Environment representing the intrusion response CMDP
    """

    def __init__(self, config: IntrusionResponseCmdpConfig):
        """
        Initializes the environment

        :param config: the environment configuration
        """
        self.config = config

        # Setup spaces
        self.action_space = gym.spaces.Discrete(len(self.config.actions))
        self.observation_space = gym.spaces.Box(low=np.int32(min(self.config.states)),
                                                high=np.int32(max(self.config.states)), dtype=np.int32, shape=(1,))

        # Initialize state
        self.t = 1
        self.s = self.config.initial_state

        # Setup traces
        self.traces: List[SimulationTrace] = []
        self.trace = SimulationTrace(simulation_env=self.config.gym_env_name)

        # Reset
        self.reset()
        super().__init__()

    def step(self, a: int) -> Tuple[int, float, bool, bool, Dict[str, Any]]:
        """
        Takes a step in the environment by executing the given action

        :param a: the action
        :return: (obs, reward, terminated, truncated, info)
        """
        done = False

        # Cost
        c = self.config.cost_tensor[self.s]

        # Sample next state
        self.s = GeneralUtil.sample_next_state(transition_tensor=self.config.transition_tensor,
                                               s=self.s, a=a, states=self.config.states)

        # Update time-step
        self.t += 1

        # Populate info dict
        info: Dict[str, Any] = {}
        info[env_constants.ENV_METRICS.STATE] = self.s
        info[env_constants.ENV_METRICS.DEFENDER_ACTION] = a
        info[env_constants.ENV_METRICS.OBSERVATION] = self.s
        info[env_constants.ENV_METRICS.TIME_STEP] = self.t
        info = self._info(info)

        # Log trace
        self.trace.defender_rewards.append(-c)
        self.trace.attacker_rewards.append(c)
        self.trace.attacker_actions.append(-1)
        self.trace.defender_actions.append(a)
        self.trace.infos.append(info)
        self.trace.states.append(self.s)
        self.trace.beliefs.append(0.0)
        self.trace.infrastructure_metrics.append(self.s)
        if not done:
            self.trace.attacker_observations.append(self.s)
            self.trace.defender_observations.append(self.s)

        return self.s, c, done, done, info

    def reset(self, seed: Union[None, int] = None, soft: bool = False, options: Union[Dict[str, Any], None] = None) \
            -> Tuple[int, Dict[str, Any]]:
        """
        Resets the environment state, this should be called whenever step() returns <done>

        :param seed: the random seed
        :param soft: boolean flag indicating whether it is a soft reset or not
        :param options: optional configuration parameters
        :return: initial observation and info
        """
        super().reset(seed=seed)
        if len(self.trace.defender_rewards) > 0:
            self.traces.append(self.trace)
        self.trace = SimulationTrace(simulation_env=self.config.simulation_env_name)
        self.t = 1
        self.s = self.config.initial_state
        self.trace.beliefs.append(0.0)
        self.trace.states.append(self.s)
        self.trace.defender_observations.append(self.s)
        self.trace.attacker_observations.append(self.s)
        info: Dict[str, Any] = {}
        return self.s, info

    def _info(self, info: Dict[str, Any]) -> Dict[str, Any]:
        """
        Adds the cumulative reward and episode length to the info dict

        :param info: the info dict to update
        :return: the updated info dict
        """
        R = 0
        for i in range(len(self.trace.defender_rewards)):
            R += self.trace.defender_rewards[i] * math.pow(self.config.discount_factor, i)
        info[env_constants.ENV_METRICS.RETURN] = sum(self.trace.defender_rewards)
        info[env_constants.ENV_METRICS.TIME_HORIZON] = len(self.trace.defender_actions)
        return info

    def render(self, mode: str = 'human'):
        """
        Renders the environment.  Supported rendering modes: (1) human; and (2) rgb_array

        :param mode: the rendering mode
        :return: True (if human mode) otherwise an rgb array
        """
        raise NotImplementedError("Rendering is not implemented for this environment")

    def get_traces(self) -> List[SimulationTrace]:
        """
        :return: the list of simulation traces
        """
        return self.traces

    def reset_traces(self) -> None:
        """
        Resets the list of traces

        :return: None
        """
        self.traces = []

    def __checkpoint_traces(self) -> None:
        """
        Checkpoints agent traces
        :return: None
        """
        ts = time.time()
        SimulationTrace.save_traces(traces_save_dir=constants.LOGGING.DEFAULT_LOG_DIR,
                                    traces=self.traces, traces_file=f"taus{ts}.json")

    def set_model(self, model) -> None:
        """
        Sets the model. Useful when using RL frameworks where the stage policy is not easy to extract

        :param model: the model
        :return: None
        """
        self.model = model

    def set_state(self, state: int) -> None:
        """
        Sets the state. Allows to simulate samples from specific states

        :param state: the state
        :return: None
        """
        self.s = state

    def manual_play(self) -> None:
        """
        An interactive loop to test the CMDP manually

        :return: None
        """
        s = self.config.initial_state
        done = False
        cumulative_costs = 0.0
        c = 0.0
        t = 1
        while True:
            raw_input = input("> ")
            raw_input = raw_input.strip()
            if raw_input == "help":
                print("Enter an action id to execute the action, "
                      "press R to reset,"
                      "press S to print the set of states"
                      "press A to print the set of actions"
                      "press s to print the current state, press A to print the actions, "
                      "press D to check if done"
                      "press C to check the cumulative costs"
                      "press H to print the history of actions")
            elif raw_input == "A":
                print(self.config.actions)
            elif raw_input == "S":
                print(self.config.states)
            elif raw_input == "s":
                print(f"s: {s}, c: {c}")
            elif raw_input == "D":
                print(done)
            elif raw_input == "C":
                print(cumulative_costs)
            elif raw_input == "R":
                print("Resetting the state")
                s = self.config.initial_state
                done = False
                cumulative_costs = 0
                c = 0.0
                t = 1
            else:
                try:
                    a = int(raw_input)
                except Exception:
                    continue
                c = self.config.cost_tensor[s]
                print(f"s: {s}, a: {a}, c: {c}")
                s = GeneralUtil.sample_next_state(transition_tensor=self.config.transition_tensor, s=s, a=a,
                                                  states=self.config.states)
                t += 1
                cumulative_costs += c
                print(f"s: {s}, c: {c}, t: {t}")
